import numpy as np 
import random
from SDC_OP2 import sdc 
from scipy import stats
from scale_noise import augment_noise_scale
import matplotlib.pyplot as plt

### Define the function for gene filter using the training dataset
def gene_filter(expr,cl):
    n_groups = cl.shape[1]  # for Lung data, n_groups=3
    dat = []
    
    for i in range(n_groups):
        dat.append(expr[cl[:,i]==1,])
    
    pvals = np.zeros(expr.shape[1])  # the pvals vector for all genes' anova test
    for i in range(expr.shape[1]):
        _,pvals[i] = stats.f_oneway(dat[0][:,i],dat[1][:,i],dat[2][:,i],dat[3][:,i])
    
    ### get the top n_genes measured by the smallest pvalue
    orders = np.argsort(-pvals)
    return orders

def mean_std_est(expr,labels):
    n_cl = labels.shape[1]  #n_cl is the number of classes
    n_gene = expr.shape[1]
    U = np.mean(expr,axis=0) # the mean expression level of each gene (mean of each column)
# estimate the mean and standard deviation within each class
    Mu = np.zeros([n_gene,n_cl]) 
    Sigma = np.zeros([n_gene,n_cl])
    for i in range(n_cl):
        Mu[:,i] =np.mean(expr[(labels[:,i]==1),],axis=0)
        Sigma[:,i] = np.sum((expr[(labels[:,i]==1),].T-Mu[:,i].reshape([n_gene,1]))*(expr[(labels[:,i]==1),].T-Mu[:,i].reshape([n_gene,1])),axis=1)

    std = np.sqrt(np.sum(Sigma,axis=1)/(expr.shape[0]-n_cl))

    ### fit a line to the data: x-U, y-std
    slope,intercept,r_value,p_value,slope_std_error=stats.linregress(U,std)
    std_h = intercept+slope*U # the predicted standard deviation

    return U,std_h

####


def augment(expr,labels,Fold=20,alpha=0.2):
    n_cell = expr.shape[0]
    # expr1 is the augmented expression matrix:
    expr1 = expr.repeat(Fold,axis=0)
    U,std_h = mean_std_est(expr,labels)
    Var = np.random.multivariate_normal(mean=np.zeros(expr.shape[1]),cov=alpha*np.diag(np.square(std_h)),size=Fold*n_cell)
    expr1 = expr1+Var
    expr1 = np.vstack((expr,expr1))
    labels1=np.tile(labels,(Fold+1,1))
    
    return expr1,labels1
    
def UQ_normalization(expr):
    expr0 = np.exp(expr)-1  # transform back to the original scale
    UQ = np.percentile(expr0,75,axis=1,keepdims=True)
    TC = np.sum(expr0,axis=1)
    TC0=TC.reshape([expr0.shape[0],1])
    expr0[TC!=0,] = (expr0[TC!=0,]/TC0[TC!=0,])*1000000
    expr1 = np.log(expr0+1)
    return expr1
#########
DATASET = "PCPG"
random.seed(40)

expr = np.loadtxt(DATASET+".txt")
cl = np.loadtxt(DATASET+"_label.txt")
## To do cross-validation, K is the number of folds that we set.
K = 5
expr = np.asarray(expr).T # Transpose the matrix, so that rows are samples and columns are genes
cl =  cl.astype(int)
N = cl.shape[0] # N is the total number of cells/ sample size
np.random.seed(200)
indx = np.random.randint(0,K,N)
## list for the test data and training data
expr_train=[]
cl_train = []
expr_test = []
cl_test = []
clist=[]
ct_list=[]
res = np.zeros(K) # res will be the accuracy of test data
res_aug_0 = np.zeros(K) # res with the data augmentation
res_aug_1 = np.zeros(K) # res with the data augmentation
res_op= np.zeros(expr.shape[0])  # res_op is vector the concordance rate for all samples
acc = 0.0  # the overall accuracy after cross-validation
acc_aug_0 = 0.0
acc_aug_1=0.0
n_genes=1500
nepoch=60
Cost = np.zeros(nepoch)
Testacc = np.zeros(nepoch)
label_hat = np.zeros(cl.shape)

zerop = np.mean(expr==0,axis=0)
expr=expr[:,zerop<0.8]


for i in range(K):
    print("The number of fold:", i+1)
    expr_train.append(expr[indx!=i,])
    cl_train.append(cl[indx!=i,])
    expr_test.append(expr[indx==i,])
    cl_test.append(cl[indx==i,])
    ## orders of negative p values for genes
    orders_t = gene_filter(expr=expr_train[i],cl=cl_train[i])
    expr_train[i] =(expr_train[i][:,orders_t<=n_genes])
    expr_test[i] = (expr_test[i][:,orders_t<=n_genes])
    
    
    res_aug_0[i],res_aug_1[i],res_op[indx==i],clist,ct_list,label_hat[indx==i,] = (sdc(expr_train=expr_train[i],labels_train=cl_train[i],expr_test=expr_test[i],labels_test=cl_test[i],lr=0.02,min_stop=0.05,epoch=nepoch,min_epoch=nepoch,aug=200,display_step=100,batch_size=24))
    
    Cost+=np.array(clist)[range(nepoch)]*expr_train[i].shape[0]
    Testacc += np.array(ct_list)[range(nepoch)]*expr_test[i].shape[0]
    acc_aug_0=acc_aug_0+res_aug_0[i]*expr_test[i].shape[0]
    acc_aug_1=acc_aug_1+res_aug_1[i]*expr_test[i].shape[0]

k = 0
lr_str = "-2e-2"
dpr = "-05"
btch = "-b-24"
wdth = ""
Cost = Cost/(expr.shape[0]*(K-1))
Testacc = Testacc/expr.shape[0]
#np.savetxt(DATASET+"/h"+str(k)+btch+lr_str+"/Cost.txt",Cost)
#np.savetxt(DATASET+"/h"+str(k)+btch+lr_str+"/Testacc.txt",Testacc)
#np.savetxt(DATASET+"/h"+str(k)+btch+lr_str+"/label_hat.txt",label_hat)
acc_aug_0 = acc_aug_0/expr.shape[0]
acc_aug_1 = acc_aug_1/expr.shape[0]
print("On the original test set the accuracy is",acc_aug_0,";The concordance rate is:",np.mean(res_op),"; Accuracy on the augmented is ",acc_aug_1)
